import pkg_resources

from rockart_examples.exceptions import (
    RockartExamplesException,
    RockartExamplesIndexError,
    RockartExamplesValueError,
)


__all__ = [
    "RockartExamplesException",
    "RockartExamplesIndexError",
    "RockartExamplesValueError",
]

try:
    __version__ = pkg_resources.get_distribution("rockart-examples").version
except pkg_resources.DistributionNotFound:
    pass



class RockartExamplesException(Exception):
    pass

class RockartExamplesIndexError(RockartExamplesException, IndexError):
    pass

class RockartExamplesValueError(RockartExamplesException, ValueError):
    pass



import curses
import itertools
import signal
import threading

from rockart import Canvas
from rockart_examples import RockartExamplesValueError
from rockart_examples.life.game import Game


def render_game(game, terminal_rows, terminal_columns):
    field_border = 1
    field_padding = 1
    cell_size = 4
    cell_border = 1

    min_canvas_size = 2 * (field_border + field_padding)
    min_terminal_rows = \
        min_canvas_size // Canvas.CELL_HEIGHT \
        + min(min_canvas_size % Canvas.CELL_HEIGHT, 1)
    min_terminal_columns = \
        min_canvas_size // Canvas.CELL_WIDTH \
        + min(min_canvas_size % Canvas.CELL_WIDTH, 1)
    if terminal_rows < min_terminal_rows:
        raise RockartExamplesValueError(
            f"`terminal_rows` must be greater or equal {min_terminal_rows}"
        )
    if terminal_columns < min_terminal_columns:
        raise RockartExamplesValueError(
            f"`terminal_columns` must be greater or equal {min_terminal_columns}"
        )

    canvas = Canvas.of_size(terminal_rows, terminal_columns - 1)

    frame_min_x = \
        max(
            canvas.width // 2 - game.columns // 2 * cell_size - (field_border + field_padding),
            0
        )
    frame_max_x = \
        min(
            canvas.width // 2 + (game.columns - game.columns // 2) * cell_size + field_padding,
            canvas.width - 1
        )
    frame_min_y = \
        max(
            canvas.height // 2 - game.rows // 2 * cell_size - (field_border + field_padding),
            0
        )
    frame_max_y = \
        min(
            canvas.height // 2 + (game.rows - game.rows // 2) * cell_size + field_padding,
            canvas.height - 1
        )
    for x in range(frame_min_x, frame_max_x + 1):
        canvas.paint(x, frame_min_y)
        canvas.paint(x, frame_max_y)
    for y in range(frame_min_y, frame_max_y + 1):
        canvas.paint(frame_min_x, y)
        canvas.paint(frame_max_x, y)

    for game_row, game_column in itertools.product(range(game.rows), range(game.columns)):
        is_alive = game.is_cell_alive(game_row, game_column)
        if not is_alive:
            continue

        cell_x = (game_column - game.columns // 2) * cell_size + (canvas.width // 2)
        cell_y = (game_row - game.rows // 2) * cell_size + (canvas.height // 2)

        for local_x in range(cell_border, cell_size - cell_border):
            for local_y in range(cell_border, cell_size - cell_border):
                x = cell_x + local_x
                if x < field_border + field_padding:
                    continue
                if x >= canvas.width - (field_border + field_padding):
                    continue

                y = cell_y + local_y
                if y < field_border + field_padding:
                    continue
                if y >= canvas.height - (field_border + field_padding):
                    continue

                canvas.paint(x, y)

    return canvas.render()


def run_gui(is_shutdown_required=None):
    if is_shutdown_required is None:
        is_shutdown_required = threading.Event()
        def handle_shutdown_signal(*args):
            is_shutdown_required.set()
        signal.signal(signal.SIGINT, handle_shutdown_signal)
        signal.signal(signal.SIGTERM, handle_shutdown_signal)

    game = Game(30, 30)
    game.initialize()

    def interact(screen):
        curses.curs_set(False)
        curses.use_default_colors()

        while not is_shutdown_required.is_set():
            rows, columns = screen.getmaxyx()
            try:
                frame = render_game(game, rows, columns)
            except RockartExamplesValueError:
                frame = ""

            screen.erase()
            try:
                screen.addstr(0, 0, frame)
            except curses.error:
                continue
            screen.refresh()

            curses.napms(50)
            game.move()
    curses.wrapper(interact)



import setuptools

def obtain_version():
    def version_scheme(version):
        return version.format_with("{tag}")

    def local_scheme(version):
        if version.exact:
            metadata = \
                version.format_choice(
                    clean_format="",
                    dirty_format="+dirty"
                )
        else:
            metadata = "+" + version.node[1:]
            metadata = \
                version.format_choice(
                    clean_format=metadata,
                    dirty_format=metadata + ".dirty"
                )
        return metadata

    return {
        "version_scheme": version_scheme,
        "local_scheme": local_scheme,
    }

with open("README.md", "r") as description_file:
    long_description = description_file.read()

setuptools.setup(
    name="rockart-examples",
    use_scm_version=obtain_version,
    description="Collection of Rockart usage examples",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords="pseudographics semigraphics braille terminal console",
    license="MIT",
    author="van-kalsing",
    author_email="kalsin@inbox.ru",
    url="https://github.com/van-kalsing/rockart-examples",
    project_urls={
        "Bug Tracker": "https://github.com/van-kalsing/rockart-examples/issues",
        "Source Code": "https://github.com/van-kalsing/rockart-examples",
    },
    zip_safe=True,
    packages=["rockart_examples", "rockart_examples.life"],
    entry_points={
        "console_scripts": [
            "rockart-life = rockart_examples.life.entry:entry",
        ],
    },
    python_requires=">=3.6.9",
    install_requires=["rockart==0.2.*"],
    setup_requires=["setuptools_scm==3.2.0"],
    classifiers=[
        "Environment :: Console",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3.6",
        "Topic :: Terminals",
    ],
)



import itertools
import random

from rockart_examples import (
    RockartExamplesException,
    RockartExamplesIndexError,
    RockartExamplesValueError
)


class Game:
    __INITIAL_LIFE_PROBABILITY = 0.1
    __NEIGHBOUR_LIFE_PROBABILITY = 0.3


    def __init__(self, rows, columns, *args, seed=None, **kwargs):
        super().__init__(*args, **kwargs)

        if rows < 0:
            raise RockartExamplesValueError("`rows` must be positive")

        if columns < 0:
            raise RockartExamplesValueError("`columns` must be positive")

        self.__is_initialized = False
        self.__epoch = None
        self.__rows = rows
        self.__columns = columns
        self.__field = [
            [False]*self.__columns for _ in range(self.__rows)
        ]
        self.__hidden_field = [
            [False]*self.__columns for _ in range(self.__rows)
        ]
        self.__random = random.Random(seed)

    @property
    def rows(self):
        return self.__rows

    @property
    def columns(self):
        return self.__columns


    @property
    def is_initialized(self):
        return self.__is_initialized

    @property
    def epoch(self):
        if not self.__is_initialized:
            raise RockartExamplesException("Initialize game first")

        return self.__epoch

    def is_cell_alive(self, row, column):
        if not self.__is_initialized:
            raise RockartExamplesException("Initialize game first")

        if not 0 <= row < self.__rows:
            raise RockartExamplesIndexError(
                f"`row` must belong to a range [0; {self.__rows - 1}])"
            )

        if not 0 <= column < self.__columns:
            raise RockartExamplesIndexError(
                f"`column` must belong to a range [0; {self.__columns - 1}])"
            )

        return self.__field[row][column]

    def initialize(self):
        if self.__is_initialized:
            raise RockartExamplesException("Game is initialized already")

        for row, column in itertools.product(range(self.__rows), range(self.__columns)):
            self.__field[row][column] = self.__random.random() < Game.__INITIAL_LIFE_PROBABILITY
        self.__epoch = 0
        self.__is_initialized = True

    def move(self):
        if not self.__is_initialized:
            raise RockartExamplesException("Initialize game first")

        for row, column in itertools.product(range(self.__rows), range(self.__columns)):
            alive_neighbours_number = 0
            for row_offset, column_offset in itertools.product([-1, 0, 1], repeat=2):
                if row_offset == column_offset == 0:
                    continue

                neighbour_row = row + row_offset
                neighbour_column = column + column_offset
                is_neighbour_internal = \
                    0 <= neighbour_row < self.__rows \
                    and 0 <= neighbour_column < self.__columns
                if is_neighbour_internal:
                    is_neighbour_alive = self.__field[neighbour_row][neighbour_column]
                else:
                    is_neighbour_alive = self.__random.random() < Game.__NEIGHBOUR_LIFE_PROBABILITY

                if is_neighbour_alive:
                    alive_neighbours_number += 1

            is_alive = self.__field[row][column]
            if is_alive:
                will_live = alive_neighbours_number in {2, 3}
            else:
                will_live = alive_neighbours_number == 3
            self.__hidden_field[row][column] = will_live
        self.__field, self.__hidden_field = self.__hidden_field, self.__field
        self.__epoch += 1



import argparse

from rockart_examples import __version__, RockartExamplesException
from rockart_examples.life.gui import run_gui


def entry():
    parser = \
        argparse.ArgumentParser(
            description="Conway's Game of Life - Rockart usage example"
        )
    if __version__ is not None:
        parser.add_argument(
            "-v", "--version",
            action="version",
            version=__version__,
            help="show version of Rockart Examples and exit"
        )
    parser.parse_args()

    try:
        run_gui()
    except RockartExamplesException as e:
        return str(e)
    except:  # noqa
        return "Unknown error occurred"
